/*
	pcd.c		single Matsushita/Panasonic CR56x CD-ROM driver for Linux

	Copyright (C) 1996 Zoltan Vorosbaranyi

	e-mail:		vbzoli@tarki.hu

	address:	H-2145 Szilasliget
			Kiss J. u. 4.
			Hungary EUROPE


	Version:	0.21 (ALPHA)


	Change log:

	0.21 
	- All the changes below for version 0.21 by Brian Candler -- 
	  B.Candler@pobox.com
	- Fixed cache bug (pcd.bufaddr value used to invalidate cache)
	- Added module param "io=0xNNN" to set base address
	- Changed to call check_region BEFORE probing
	- Modified to allow 512-byte blocks (used by Macintosh HFS)
	- Made some used-once functions inline

	0.20
	- Sound Blaster interface support added
	  (Zoltan Pataki (zpataki@amon.omgk.hu) was kind to me lending 
	  his SoundBlaster card, thank you Zoli!)
	- autodetection added

	0.18
	- pcd_search_end now determines the size of the disk
	  (searches for the last readable block)
	  blk_size is now correctly set, dd won't return
	  with I/O errors.
	- minor changes in pcd_readcmd, pcd_init

	0.17:
	- tested with kernel 2.0.0 and kerneld
	- implemented new ioctl's: EJECT_SW, RESET, VOLREAD, CLOSETRAY
	- resetless initialization (useful for command line audio
	  players and kerneld)
	- new option: CLOSE_TRAY_ON_OPEN
	- compiles for the 2.0.x kernels by default (no need for -DNEW)
	
	Single Matsushita/Panasonic CR-56x CDROM driver.  It's tested
	with linux 1.2.13 ELF, pre2.0.7 and 2.0 as a module with a 
	no-sound interface card and GUS MAX soundcard.  
	Works fine with WorkMan and with data CD's. 
	There's autodetection of base ports, but you can
	set base address of your interface (PCD_BASEPORT)
	before compiling, or you can set the base port
	via the io=0xNNN command line parameter.
	If you want to use this driver with old (1.2.xx)
	kernels please insert -DOLD into the C compiler flags.
	If you use kerneld, then insert the following
	line in your /etc/conf.modules:
	
		alias block-major-25 pcd
	
	
	NOTE: Driver is in alpha stage.  Use sbpcd if you want 
	a stable driver.


	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/blkdev.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/cdrom.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/timer.h>

#include <asm/io.h>

#define MAJOR_NR	MATSUSHITA_CDROM_MAJOR

#ifndef OLD
#include <linux/blk.h>
#else
#include "/usr/src/linux/drivers/block/blk.h"
#endif

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#ifdef OLD
#ifndef CONFIG_MODVERSIONS
char kernel_version[] = UTS_RELEASE;
#endif
#endif
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

/*
 ***	User servicable part begins	***
 */

/*
 * if AUTOHUNT is defined, the driver will try to detect the base port
 * (interface type is always auto-detected)
 */
#define AUTOHUNT

/*
 * base port of the interface card
 * this value is tried first during autodetection
 */
#define PCD_BASEPORT	0x340

/*
 * if PCD_LOCK is defined, the driver tries to keep the tray locked
 */
#define PCD_LOCK

/*
 * if RESET_ON_INIT is defined, the driver will reset the
 * drive during the initialization
 */
/*#define RESET_ON_INIT*/

/*
 * close tray on device open?
 */
#define CLOSE_TRAY_ON_OPEN

/*
 * invalidate buffer pages on close?
 * (if not defined it will invalidate buffer pages during
 * the pcd_open if media had changed)
 */
#define INVALIDATE_BUF_ON_CLOSE

/*
 ***	End of user servicable part	***
 */

int pcd_ports[] = 
{ 
#ifdef PCD_BASEPORT
	PCD_BASEPORT,
#endif
	0x340, 0x320, 0x330, 0x360, 0x230, 0x250, 0x270, 0x290, 
	0x300, 0x310, 0x330, 0x350, 0x370, 0x240, 0x260, 
	0x630, 0x650, 0x670, 0x690, 
	0
};

/*
 * Buffer size in frames
 */
#define PCD_BUFFER_SIZE	4
#define PCD_READ_AHEAD	16

/*
 * time out constants
 */
#define BUSY_TO		1024
#define DRIVE_SLEEP_TO	(2 * HZ)
#define DISK_SLEEP_TO	(5 * HZ)
#define TRAY_SLEEP_TO	(12 * HZ)
#define READ_SLEEP_TO	(4 * HZ)
/* fast time out value -- useful when probing the device */
#define FAST_TO		HZ
#define TIMEOUT		1
/* don't change! */
#define	OK		0

/*
 * delay after reset command
 */
#define RESET_DELAY	(3 * HZ)	/* 3 s */

/*
 * retry constants for pcd_docmd
 */
#define RETRY_DELAY	(HZ / 2)	/* 0.5 s */
#define RETRY_N		6

/*
 * retry constants and macro for pcd_open
 */
#define OPEN_RETRY_N	10
#define OPEN_DELAY	HZ		/* 1 s */

/*
 * constants for pcd_close
 */
#define CLOSETRAY_RETRY	20
#define CLOSETRAY_DELAY	HZ

#define OPEN_RETRYABLE(E)	\
	((E) == ERR_TRAYOUT || (E) == ERR_NODISK || (E) == ERR_BUSY || (E) == ERR_NOTREADY)

/*
 * command classes
 */
#define DRIVECMDCLASS	1	/* cmd involving only the drive */
#define TRAYCMDCLASS	2	/* long operations with the tray (open, close) */
#define DISKCMDCLASS	3	/* audio or data operations ok */
#define DATACMDCLASS	4	/* data operations ok */
#define AUDIOCMDCLASS	5	/* audio operations ok */

/*
 * bits in pcd.ready
 */
#define RDY_DATA	0x02	/* data arrived */
#define RDY_REPLY	0x04	/* cmd reply arrived */

/*
 * error codes
 */
#define ERR_NOERROR	0x00
#define ERR_SR_RETRY	0x01
#define ERR_SR_EC	0x02
#define ERR_NOTREADY	0x03
#define ERR_HR		0x05
#define ERR_SEEKNCOMPL	0x06
#define ERR_DISKCHANGED	0x11
#define	ERR_RESET	0x12
#define ERR_TIMEOUT	(-1)
#define ERR_TRAYOUT	(-2)
#define ERR_NODISK	(-3)
#define ERR_AUDIOPLAY	(-4)
#define ERR_BUSY	(-5)

/*
 * command codes
 */
#define CMD_STATUS	0x05
#define CMD_EJECT	0x06
#define CMD_CLOSE	0x07
#define CMD_STOP	0x08
#define CMD_SETVOLUME	0x09
#define CMD_LOCK	0x0c
#define CMD_PAUSE_RES	0x0d
#define CMD_PLAYMSF	0x0e
#define CMD_PLAYTI	0x0f
#define CMD_READ	0x10
#define CMD_GETERRCODE	0x82
#define CMD_VERSION	0x83
#define CMD_READVOLUME	0x84
#define CMD_READSUBQ	0x87
#define CMD_DISKINFO	0x8b
#define CMD_READTOC	0x8c
#define CMD_MULTI	0x8d

/*
 * values for CMD_SETVOLUME, CMD_READVOLUME
 */
#define MODE_VOL	5
#define VOL_SIZE	5
#define VOL_VALUE	2

/*
 * status bits
 */
#define STAT_READY	0x01
#define STAT_PLAY	0x08
#define STAT_ERROR	0x10
#define STAT_DISK	0x40
#define STAT_TRAY	0x80

/*
 * for GETERRCODE cmd
 */
#define ERRSIZE		8
#define ERRBYTE		2

/*
 * for CMD_SUBQ command
 */
#define SUBQ_TYPE	1
#define SUBQ_TRACK	2
#define SUBQ_INDEX	3
#define SUBQ_ABS_M	4
#define SUBQ_ABS_S	5
#define SUBQ_ABS_F	6
#define SUBQ_REL_M	7
#define SUBQ_REL_S	8
#define SUBQ_REL_F	9
#define SUBQ_SIZE	11

/*
 * for READTOC command
 */
#define TOCBUF_SIZE	8
#define TOC_TYPE	1
#define TOC_NUM		2
#define TOC_DATAMODE	3
#define TOC_M		4
#define TOC_S		5
#define TOC_F		6
#define	TOC_SIZE	120

/*
 * for DISKINFO command
 */
#define DINFOBUF_SIZE	6
#define DINFO_TYPE	0
#define DINFO_STARTTRK	1
#define DINFO_ENDTRK	2
#define DINFO_M		3
#define DINFO_S		4
#define DINFO_F		5

/*
 * for CMD_MULTI 
 */
#define MULTI_SIZE	6
#define MULTI_FLAG	0x80
#define MULTI_B		0
#define MULTI_M		1
#define MULTI_S		2
#define MULTI_F		3

/*
 * 2 s data offset in frames
 */
#define DATA_OFFSET	CD_BLOCK_OFFSET

/*
 * Data track ctrl value in the toc
 */
#define DATA_CTRL	CDROM_DATA_TRACK

/*
 * for CMD_PAUSE_RES command
 */
#define P_RESUME	0x80

/*
 * Size of a version string
 * returned by VERSION_CMD
 * and displayed
 */
#define VERSION_SIZE	10
#define VERS_DISP_SIZE	13
#define VERS_MODEL_LEN	6

/*
 *	Selection values for the SB interface
 */
#define PCD_SEL_DATA	1
#define PCD_SEL_STATUS	0

/*
 *	Interface type values (pcd.if_type)
 */
#define PCD_SB_IF	0
#define PCD_NORMAL_IF	1

/*
 * Kernel sector and CD block sizes
 */
#define SECTOR_SIZE	512
#define CDBLOCK_SIZE	CD_FRAMESIZE

/*
 * Values for pcd_search_end
 */
#define SEARCH_BACK	75
#define SEARCH_N	10

#define OUTB		outb
#define INB		inb

/*
 * error messages
 *   > 0 returned by the drive
 *   < 0 other errors
 */
char *errstr[] = {
	"Drive busy",				/* -5 */
	"Audio playing",
	"No disk in the drive",
	"Tray is out",
	"Time out (drive not responding)",
	"No error (driver error if displayed!)",/* 0 */
	"Soft read error after retry",
	"Soft read error after error correction",
	"Drive not ready",
	"Cannot read TOC",
	"Hard read error",
	"Seek didn't complete",
	"Tracking servo failure",
	"Drive RAM error",
	"Self-test failed",
	"Focusing servo failure",
	"Spindle servo failure",
	"Data path failure",
	"Illegal logical block address",
	"Illegal field in CDB",
	"End of user encountered on this track ?",
	"Illegal data mode for this track",
	"Media changed",
	"Power-on or reset occured",
	"Drive ROM failure",
	"Illegal drive command from the host",
	"Disk removed during operation",
	"Drive hardware error",
	"Illegal request from host"
};

#define ERR_N		(sizeof(errstr) / sizeof(char *))
#define ERROFFS		5		/* error no. offset */
#define	ERRSTR(EC)	(errstr[(EC) + ERROFFS])

#define PCD_CMD_C(F, EC)				\
	if ((EC = pcd_ ## F ())) {			\
		perr(EC);				\
		return -EIO;				\
	}

/*
 * minute-second-frame address type
 */
struct MSF {
	u_char m, s, f;
};

struct TOC_ENTRY {
	u_char num;
	u_char adr:4;
	u_char ctrl:4;
	u_char datamode;
	struct MSF start;
};

struct TOC {
	u_char type;
	u_char start_track;
	u_char end_track;
	int datasize;		/* size of the data track */
	struct MSF size;	/* disk size */
	struct TOC_ENTRY entry[TOC_SIZE];
};

struct SUBQ {
	u_char adr:4;
	u_char ctrl:4;
	u_char track;
	u_char index;
	struct MSF absaddr;
	struct MSF reladdr;
};

struct MULTI {
	int flag;
	struct MSF base;
};

struct PCD_S {
	/*
	 * interface ports
	 */
	int base;		/* base port */
	int ready;
	int data;		/* data port */
	int reset;		/* reset port */
	int select;		/* selection port (for SB) */

	int if_type;		/* interface type SB or normal */

	int present;		/* drive present */
	u_char status;		/* last drive status */
	u_char prev_status;	/* previous status */
	u_char ready_value;	/* last seen ready value */
	int open_count;
	
	/*
	 * change flags; each flag is set when ``media changed'' status returned
	 */
	int changed;		/* disk changed (reset by close) */
	int change_flag;	/* flag for check_pcd_change (reset by check_pcd_change) */
	int update_toc;		/* toc needs updating (reset by pcd_readtoc) */

	int paused;		/* audio play paused */
	int unlocked;		/* drive is unlocked (after eject command) */

	int eject_sw;		/* eject on close */

	struct TOC toc;		/* table of contents */
	struct SUBQ subq;	/* subchannel info */
	struct MULTI multi;	/* multisession info (?) */

	int bufaddr;		/* address of first block in the buffer */
	char buffer[PCD_BUFFER_SIZE * CDBLOCK_SIZE]; /* block buffer */
} pcd;

struct PCD_CMD {
	u_char b1, b2, b3, b4, b5, b6;
};

#define MAX_PCD 	4	/* for future use (only 1 drive at present) */
static int pcd_blk_size[MAX_PCD] = { 0, 0, 0, 0 };
static int pcd_blocksizes[MAX_PCD] = { 0, 0, 0, 0 };
static int io[MAX_PCD] = { 0, 0, 0, 0 };	/* base addr from insmod */

#ifdef PCD_STAT
#define MAXTIMEOUT	10

struct PCD_STATISTICS {
	unsigned busyloop[BUSY_TO];
	unsigned sleeploop[DISK_SLEEP_TO];
	unsigned sleeptime[DISK_SLEEP_TO];
	int timeout_index;
	unsigned timeout[MAXTIMEOUT];
	unsigned cmdretry[RETRY_N];
	unsigned totalreads;
	unsigned readcachemisses;
	unsigned readcachehits;
	unsigned delayloop;
} stat;
#endif

static void pcd_delay(unsigned int t);
static int wait_data(u_char ready_bit, int sleep_to);
int pcd_ioctl(struct inode *ip, struct file *fp,
              unsigned int cmd, unsigned long arg);
int pcd_open(struct inode *ip, struct file *fp);
void pcd_release(struct inode *inode, struct file *fp);
int check_pcd_change(dev_t dev);
static int pcd_docmd(int class, int cmd, struct PCD_CMD *cs, u_char *rbuf, int rn);

#if 0
static void pcd_timerproc(unsigned long data);
static struct timer_list pcd_timer = { NULL, NULL, 0, 0, pcd_timerproc };
static struct wait_queue *pcd_waitqueue = NULL;
#endif

static struct file_operations pcd_fops = {
	NULL,			/* lseek */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir */
	NULL,			/* select */
	pcd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	pcd_open,		/* open */
	pcd_release,		/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	check_pcd_change,	/* media change */
	NULL			/* revalidate */
};

/*
 * block <-> minute/second/frame conversion 
 */
static int msf2block(u_char m, u_char s, u_char f)
{
	return m * (CD_SECS * CD_FRAMES) + s * CD_FRAMES + f - DATA_OFFSET;
}

static void block2msf(u_char *m, u_char *s, u_char *f, int block)
{
	block += DATA_OFFSET;
	*f = block % CD_FRAMES;
	block = block / CD_FRAMES;
	*s = block % CD_SECS;
	*m = block / CD_SECS;
}

#if 0
static int msf_eq(u_char m1, u_char s1, u_char f1, u_char m2, u_char s2, u_char f2)
{
	Return m1 == m2 && s1 == s2 && f1 == f2;
}
#endif

static inline int msf_less(u_char m1, u_char s1, u_char f1, u_char m2, u_char s2, u_char f2)
{
	if (m1 == m2) {
		if (s1 == s2)
			return f1 < f2;
		else
			return s1 < s2;
	} else
		return m1 < m2;
}

static struct PCD_CMD *mkc(struct PCD_CMD *cs, int p1, int p2, int p3, int p4, int p5, int p6)
{
	cs->b1 = p1;
	cs->b2 = p2;
	cs->b3 = p3;
	cs->b4 = p4;
	cs->b5 = p5;
	cs->b6 = p6;
	return cs;
}

static void clear_cmd(struct PCD_CMD *cs)
{
	cs->b1 = 0;
	cs->b2 = 0;
	cs->b3 = 0;
	cs->b4 = 0;
	cs->b5 = 0;
	cs->b6 = 0;
}

static void pcd_cmd(int cmd, struct PCD_CMD *cs)
{
	OUTB(cmd, pcd.base);
	OUTB(cs->b1, pcd.base);
	OUTB(cs->b2, pcd.base);
	OUTB(cs->b3, pcd.base);
	OUTB(cs->b4, pcd.base);
	OUTB(cs->b5, pcd.base);
	OUTB(cs->b6, pcd.base);
}

static void pcd_cmd0(int cmd)
{
	int i;

	OUTB(cmd, pcd.base);
	for (i = 0; i < 6; ++i)
		OUTB(0, pcd.base);
}

static void perr(int ec)
{
	printk("pcd: Error: %s.\n", ERRSTR(ec));
}

/*
 * read error code
 */
static inline int pcd_errcode(void)
{
	int i;
	char errbuf[ERRSIZE];

	pcd_cmd0(CMD_GETERRCODE);
	if (OK != wait_data(RDY_REPLY, DRIVE_SLEEP_TO))
		return ERR_TIMEOUT;
	for (i = 0; i < ERRSIZE; ++i)
		errbuf[i] = INB(pcd.base);
	if (errbuf[ERRBYTE] == ERR_DISKCHANGED) {
		pcd.changed = 1;
		pcd.change_flag = 1;
		pcd.update_toc = 1;
	}
	pcd.status = INB(pcd.base);
	return errbuf[ERRBYTE];
}

/*
 * read status u_char
 * 
 * returned error code depends on command class
 */
static int pcd_readstatus(int class)
{
	int errcode = 0;

	if ((pcd.status = INB(pcd.base)) & STAT_ERROR)
		errcode = pcd_errcode();
	if (class != DRIVECMDCLASS && class != TRAYCMDCLASS) {
		if (!(pcd.status & STAT_TRAY))
			return ERR_TRAYOUT;
		if (!(pcd.status & STAT_DISK))
			return ERR_NODISK;
		if (!(pcd.status & STAT_READY))
			return ERR_BUSY;
		if (class == DATACMDCLASS && pcd.status & STAT_PLAY)
			return ERR_AUDIOPLAY;
	}
	return errcode;
}

static int pcd_reset(void)
{
	OUTB(0, pcd.reset);
	pcd_delay(RESET_DELAY);
	pcd_cmd0(CMD_STATUS);
	if (OK != wait_data(RDY_REPLY, FAST_TO))
		return 0;
	if (ERR_RESET != pcd_readstatus(DRIVECMDCLASS))
		return 0;
	return 1;
}

static inline int pcd_version(char *s)
{
	int i, e;
	char ver[VERSION_SIZE];

	e = pcd_docmd(DRIVECMDCLASS, CMD_VERSION, NULL, ver, VERSION_SIZE);
	if (e)
		return e;
	for (i = 0; i < VERS_MODEL_LEN; ++i)
		*s++ = ver[i];
	*s++ = ',';
	*s++ = ' ';
	for (; i < VERSION_SIZE; ++i)
		*s ++ = ver[i];
	*s = 0;
	return 0;
}

/*
 * Determine interface type, and return 0 on success
 */
static inline int pcd_if_type(void)
{
	u_char c;

	pcd_cmd0(CMD_STATUS);
	if (OK != wait_data(RDY_REPLY, DRIVE_SLEEP_TO))
		return 1;
	OUTB(PCD_SEL_DATA, pcd.select);
	(void)INB(pcd.base);
	c = INB(pcd.ready) & RDY_REPLY;
	OUTB(PCD_SEL_STATUS, pcd.select);
	(void)INB(pcd.base);
	if (c) {
		pcd.if_type = PCD_NORMAL_IF;
		pcd.data = pcd.base + 2;
	} else {
		pcd.if_type = PCD_SB_IF;
		pcd.data = pcd.base;
	}
#ifdef PCD_DEBUG
	printk("pcd: if_type: %d\n", pcd.if_type);
#endif
	return 0;
}

/*
 * check the drive (it needs further development!)
 */
static int pcd_probe(int port, char *version)
{
	int e;

	if (check_region(port, 4)) {
		printk("pcd: Ports 0x%x-0x%x already in use.\n", port, port + 3);
		return 0;
	}
	pcd.base = port;
	pcd.ready = pcd.base + 1;
	pcd.select = pcd.base + 1;
	pcd.reset = pcd.base + 2;
#ifdef RESET_ON_INIT
	e = pcd_reset();
	if (e == 0)
		return 0;
#else
	/* ignore the first status... */
	(void)pcd_readstatus(DRIVECMDCLASS);
	e = pcd_readstatus(DRIVECMDCLASS);
	if (e && e != ERR_RESET && e != ERR_DISKCHANGED)
		return 0;
#endif
	if(pcd_if_type())
		return 0;
	if(pcd_version(version))
		return 0;
	pcd.prev_status = pcd.status;
	return 1;
}

/*
 * if there's a response from the drive read it
 */
static inline void pcd_readresp(u_char *buf, int n)
{
	int i;

	for (i = 0; i < n; ++i)
		buf[i] = INB(pcd.base);
}

/*
 * determine time out value
 */
static inline int timeout_value(int class)
{
	switch (class) {
	case DRIVECMDCLASS:
		return DRIVE_SLEEP_TO;
	case TRAYCMDCLASS:
		return TRAY_SLEEP_TO;
	default:
		return DISK_SLEEP_TO;
	}
}

/* 
 * retry command?
 */
static int retry_cmd(int e, int class)
{
	switch (e) {
	/* retryable errors: */
	case ERR_SR_RETRY:
	case ERR_SR_EC:
	case ERR_SEEKNCOMPL:
		return 1;
	case ERR_BUSY:
	case ERR_NOTREADY:
	case ERR_DISKCHANGED:	/* is this a good idea? */
		if (class == AUDIOCMDCLASS)
			return 1;
		/* fall through */
	default:
		return 0;
	}
}

/*
 * issue a command, read response, retry
 */
static int pcd_docmd(int class, int cmd, struct PCD_CMD *cs, u_char *rbuf, int rn)
{
	int i, e;

	pcd.prev_status = pcd.status;
	for (i = 1; i <= RETRY_N; ++i) {
		/* issue command */
		if (cmd == CMD_STOP)
			OUTB(cmd, pcd.base);
		else {
			if (cs)
				pcd_cmd(cmd, cs);
			else
				pcd_cmd0(cmd);
		}
		/* wait for status reply */
		if (wait_data(RDY_REPLY, timeout_value(class)) == TIMEOUT) {
			e = ERR_TIMEOUT;
			break;
		}
		/* read command response if it is required */
		if (rbuf)
			pcd_readresp(rbuf, rn);
		/* read status (it's the last u_char in the response) */
		e = pcd_readstatus(class);
		if (e == 0) {
#ifdef PCD_STAT
			if (i > 1)
				++stat.cmdretry[i - 1];
#endif
			return 0;
		}
		if (retry_cmd(e, class))
			/* increase delay time between retries */
			pcd_delay(i * RETRY_DELAY);
		else
			break;
	}
#ifdef PCD_DEBUG
	printk("pcd: Error: %s (0x%02x), retries = %d.\n", ERRSTR(e), e, i - 1);
	if (cs)
		printk("pcd: cmd = %02x, %d %d %d %d %d %d\n", cmd, cs->b1, cs->b2, cs->b3, cs->b4, cs->b5, cs->b6);
	else
		printk("pcd: cmd = %02x, 0 0 0 0 0 0\n", cmd);
#endif
#ifdef PCD_STAT
	if (i > 1)
		++stat.cmdretry[i - 1];
#endif
	return e;
}

/*
 * read data -- n * CDBLOCK_SIZE u_chars 
 */
static inline void read_data(int n, char *buf)
{
	int j;
	char *p;

	for (j = 0, p = buf; j < n; ++j, p += CDBLOCK_SIZE) {
		if (wait_data(RDY_DATA | RDY_REPLY, DISK_SLEEP_TO) == TIMEOUT)
			return;
		/* there is an error if reply came first */
		if ((pcd.ready_value & RDY_REPLY) == 0)
			return;
		if (pcd.if_type == PCD_SB_IF)
			OUTB(PCD_SEL_DATA, pcd.select);
		insb(pcd.data, p, CDBLOCK_SIZE);
		if (pcd.if_type == PCD_SB_IF)
			OUTB(PCD_SEL_STATUS, pcd.select);
	}
}

/*
 * read n blocks from the drive
 */
static int pcd_readcmd(u_char m, u_char s, u_char f, char *buf, int n)
{
	struct PCD_CMD cs;
	int i, e;
#ifdef PCD_DEBUG
	printk("pcd: read %d blocks m/s/f = %d/%d/%d\n", n, m, s, f);
#endif
	pcd.prev_status = pcd.status;
	for (i = 1; i <= RETRY_N; ++i) {
		/* issue read command */
		pcd_cmd(CMD_READ, mkc(&cs, m, s, f, 0, 0, n));
		/* read data -- n * CDBLOCK_SIZE u_chars */
		read_data(n, buf);
		/* wait for status reply */
		if (wait_data(RDY_REPLY, READ_SLEEP_TO) == TIMEOUT) {
			e = ERR_TIMEOUT;
			break;
		}
		e = pcd_readstatus(DATACMDCLASS);
		if (e == 0) {
#ifdef PCD_STAT
			if (i > 1)
				++stat.cmdretry[i - 1];
#endif
			return 0;
		}
		if (retry_cmd(e, DATACMDCLASS))
			/* increase delay time between retries */
			pcd_delay(i * RETRY_DELAY);
		else
			break;
	}
#ifdef PCD_DEBUG
	printk("pcd: (readcmd) Error: %s (0x%02x), retries = %d.\n", ERRSTR(e), e, i - 1);
	printk("pcd: cmd = %02x, %d %d %d %d %d %d\n", CMD_READ, cs.b1, cs.b2, cs.b3, cs.b4, cs.b5, cs.b6);
#endif
#ifdef PCD_STAT
	if (i > 1)
		++stat.cmdretry[i - 1];
#endif
	return e;
}

static inline int pcd_getstatd(void)
{
	return pcd_docmd(DISKCMDCLASS, CMD_STATUS, NULL, NULL, 0);
}

static int pcd_getstat(void)
{
	return pcd_docmd(DRIVECMDCLASS, CMD_STATUS, NULL, NULL, 0);
}

static inline int pcd_stop(void)
{
	return pcd_docmd(DRIVECMDCLASS, CMD_STOP, NULL, NULL, 0);
}

static inline int pcd_pause(void)
{
	int e;

	if ((pcd.status & STAT_PLAY) && !pcd.paused) {
		e = pcd_docmd(AUDIOCMDCLASS, CMD_PAUSE_RES, NULL, NULL, 0);
		if (e)
			return e;
		pcd.paused = 1;
	}
	return 0;
}

static inline int pcd_resume(void)
{
	int e;
	struct PCD_CMD cmd;

	if ((pcd.status & STAT_PLAY) && pcd.paused) {
		clear_cmd(&cmd);
		cmd.b1 = P_RESUME;
		e = pcd_docmd(AUDIOCMDCLASS, CMD_PAUSE_RES, &cmd, NULL, 0);
		pcd.paused = 0;
		if (e)
			return e;
	}
	return 0;
}

static inline int pcd_playti(u_char from_t, u_char from_i, u_char to_t, u_char to_i)
{
	struct PCD_CMD cmd;

	pcd.paused = 0;
	return pcd_docmd(AUDIOCMDCLASS, CMD_PLAYTI, mkc(&cmd, from_t, from_i, to_t, to_i, 0, 0), NULL, 0);
}

static inline int pcd_playmsf(u_char from_m, u_char from_s, u_char from_f, u_char to_m, u_char to_s, u_char to_f)
{
	struct PCD_CMD cmd;
	
	pcd.paused = 0;
	return pcd_docmd(AUDIOCMDCLASS, CMD_PLAYMSF, mkc(&cmd, from_m, from_s, from_f, to_m, to_s, to_f), NULL, 0);
}

static int pcd_lock(void)
{
	struct PCD_CMD cmd;

	clear_cmd(&cmd);
	cmd.b1 = 1;
	return pcd_docmd(DRIVECMDCLASS, CMD_LOCK, &cmd, NULL, 0);
}

static int pcd_unlock(void)
{
	return pcd_docmd(DRIVECMDCLASS, CMD_LOCK, NULL, NULL, 0);
}

static inline int pcd_eject(void)
{
	return pcd_docmd(TRAYCMDCLASS, CMD_EJECT, NULL, NULL, 0);
}

static int pcd_close(int wait)
{
	int e, i;

	e = pcd_docmd(TRAYCMDCLASS, CMD_CLOSE, NULL, NULL, 0);
	if (e)
		return e;
	if (wait) {
		for (i = 0; i < CLOSETRAY_RETRY; ++i) {
			e = pcd_getstat();
			if (pcd.status & STAT_TRAY)
				break;
			pcd_delay(CLOSETRAY_DELAY);
		}
	}
	return 0;
}

static inline int pcd_volume(u_char vol)
{
	struct PCD_CMD cmd;

	return pcd_docmd(AUDIOCMDCLASS, CMD_SETVOLUME, mkc(&cmd, MODE_VOL, 0, 1, vol, 2, vol), NULL, 0);
}

static inline int pcd_readvolume(u_char *volp)
{
	int e;
	struct PCD_CMD cmd;
	u_char buf[VOL_SIZE];

	clear_cmd(&cmd);
	cmd.b1 = MODE_VOL;
	e = pcd_docmd(AUDIOCMDCLASS, CMD_READVOLUME, &cmd, buf, VOL_SIZE);
	if (e)
		return e;
	*volp = buf[VOL_VALUE];
	return 0;
}

static inline int pcd_readsubq(void)
{
	int e;
	u_char buf[SUBQ_SIZE];

	e = pcd_docmd(DISKCMDCLASS, CMD_READSUBQ, NULL, buf, SUBQ_SIZE);
	if (e)
		return e;
	pcd.subq.adr = (buf[SUBQ_TYPE] & 0xf0) >> 4;
	pcd.subq.ctrl = buf[SUBQ_TYPE] & 0x0f;
	pcd.subq.track = buf[SUBQ_TRACK];
	pcd.subq.index = buf[SUBQ_INDEX];
	pcd.subq.absaddr.m = buf[SUBQ_ABS_M];
	pcd.subq.absaddr.s = buf[SUBQ_ABS_S];
	pcd.subq.absaddr.f = buf[SUBQ_ABS_F];
	pcd.subq.reladdr.m = buf[SUBQ_REL_M];
	pcd.subq.reladdr.s = buf[SUBQ_REL_S];
	pcd.subq.reladdr.f = buf[SUBQ_REL_F];
	return 0;
}

/*
 * read multisession info
 */
static inline int pcd_multi(void)
{
	int e;
	u_char buf[MULTI_SIZE];

	e = pcd_docmd(DISKCMDCLASS, CMD_MULTI, NULL, buf, MULTI_SIZE);
	if (e)
		return e;
	if (buf[MULTI_B] & MULTI_FLAG) {
		pcd.multi.flag = 1;
		pcd.multi.base.m = buf[MULTI_M];
		pcd.multi.base.s = buf[MULTI_S];
		pcd.multi.base.f = buf[MULTI_F];
	} else
		pcd.multi.flag = 0;
	return 0;
}

/*
 * read toc header
 */
static inline int pcd_diskinfo(void)
{
	int ec;
	u_char buf[DINFOBUF_SIZE];

	if ((ec = pcd_docmd(DISKCMDCLASS, CMD_DISKINFO, NULL, buf, DINFOBUF_SIZE)))
		return ec;
	pcd.toc.type = buf[DINFO_TYPE];
	pcd.toc.start_track = buf[DINFO_STARTTRK];
	pcd.toc.end_track = buf[DINFO_ENDTRK];
	pcd.toc.size.m = buf[DINFO_M];
	pcd.toc.size.s = buf[DINFO_S];
	pcd.toc.size.f = buf[DINFO_F];
#ifdef PCD_DEBUG
	printk("pcd: Diskinfo: type:%x, start:%d, end:%d, m/s/f:%d/%d/%d\n",
		pcd.toc.type,
		pcd.toc.start_track,
		pcd.toc.end_track,
		pcd.toc.size.m,
		pcd.toc.size.s,
		pcd.toc.size.f);
#endif
	return 0;
}

/*
 * This routine tries to search for the end of the data track.
 * Returns the real size for the data track.
 */
static inline int pcd_search_end(int size)
{
	u_char m, s, f;
	int block, e, j;

	/* step back SEARCH_BACK blocks until a good block is reached */
	for (block = size - 1, j = 0, e = -1; block >= 0 && j < SEARCH_N; block -= SEARCH_BACK, ++j) {
		block2msf(&m, &s, &f, block);
		e = pcd_readcmd(m, s, f, pcd.buffer, 1);
#ifdef PCD_DEBUG
		printk("pcd: search: m/s/f=%d/%d/%d, %s\n", m, s, f, e ? "bad" : "good");
#endif
		if (!e)
			break;
	}
	/* give up if good block not found */
	if (e)
		return 0;
	/* now search forward for the first readable block */
	for (++block; !e && block < size; ++block) {
		block2msf(&m, &s, &f, block);
		e = pcd_readcmd(m, s, f, pcd.buffer, 1);
#ifdef PCD_DEBUG
		printk("pcd: search: m/s/f=%d/%d/%d, %s\n", m, s, f, e ? "bad" : "good");
#endif
	}
	/* return the size (not the last good block) */
	return block - 1;
}

/*
 * read table of contents
 */
static int pcd_readtoc(void)
{
	int e, i;
	struct PCD_CMD cmd;
	struct MSF msf;
	u_char buf[TOCBUF_SIZE];

	/* read toc header */
	if ((e = pcd_diskinfo()))
		return e;
	/* read multisession info */
	if ((e = pcd_multi()))
		return e;
	clear_cmd(&cmd);
	for (i = pcd.toc.start_track; i <= pcd.toc.end_track; ++i) {
		cmd.b2 = i;
		e = pcd_docmd(DISKCMDCLASS, CMD_READTOC, &cmd, buf, TOCBUF_SIZE);
		if (e)
			return e;
		pcd.toc.entry[i].num = buf[TOC_NUM];
		pcd.toc.entry[i].adr = (buf[TOC_TYPE] & 0xf0) >> 4;
		pcd.toc.entry[i].ctrl = buf[TOC_TYPE] & 0x0f;
		pcd.toc.entry[i].datamode = buf[TOC_DATAMODE];
		pcd.toc.entry[i].start.m = buf[TOC_M];
		pcd.toc.entry[i].start.s = buf[TOC_S];
		pcd.toc.entry[i].start.f = buf[TOC_F];
	}
	/* now fill the "lead-out" record */
	pcd.toc.entry[i].num = CDROM_LEADOUT;
	pcd.toc.entry[i].datamode = 0;
	pcd.toc.entry[i].adr = 0;
	pcd.toc.entry[i].ctrl = 0;
	pcd.toc.entry[i].start = pcd.toc.size;
	/* determine size of the data track */
	if (pcd.toc.entry[pcd.toc.start_track].ctrl & DATA_CTRL) {
		msf = pcd.toc.entry[pcd.toc.start_track + 1].start;
		pcd.toc.datasize = msf2block(msf.m, msf.s, msf.f);
	} else
		pcd.toc.datasize = 0;
	pcd.update_toc = 0;
	pcd.toc.datasize = pcd_search_end(pcd.toc.datasize);
	/* device size in 1K blocks */
	blk_size[MAJOR_NR][0] = pcd.toc.datasize * 2;
#ifdef PCD_DEBUG
	for (i = pcd.toc.start_track; i <= pcd.toc.end_track + 1; ++i) {
		printk("pcd:   %d: num:%d, adr:%x, ctrl:%x, form:%x, m/s/f:%d/%d/%d\n",
			i,
			pcd.toc.entry[i].num,
			pcd.toc.entry[i].adr,
			pcd.toc.entry[i].ctrl,
			pcd.toc.entry[i].datamode,
			pcd.toc.entry[i].start.m,
			pcd.toc.entry[i].start.s,
			pcd.toc.entry[i].start.f);
	}
	printk("pcd: datasize = %d\n", pcd.toc.datasize);
	printk("pcd: multi = %d", pcd.multi.flag);
	if (pcd.multi.flag)
		printk(", base = %d/%d/%d\n", 
		       pcd.multi.base.m,
		       pcd.multi.base.s,
		       pcd.multi.base.f);
	else
		printk("\n");
#endif
	return 0;
}

static inline int pcd_hunt(char *version)
{
	int i;

	for (i = 0; pcd_ports[i]; ++i) {
		if (pcd_probe(pcd_ports[i], version))
			return 1;
		else
			printk("pcd: no Panasonic CD-ROM found on address 0x%x\n", pcd_ports[i]);
	}
	return 0;
}

#ifdef MODULE
int init_module(void)
#else
unsigned long pcd_init(unsigned long mem_start, unsigned long mem_end)
#endif
{
	int i;
	char version[VERS_DISP_SIZE];

#ifdef MODULE
	if (io[0])
		i = pcd_probe(io[0], version);
	else
#endif
#ifndef AUTOHUNT
	i = pcd_probe(PCD_BASEPORT, version);
#else
	i = pcd_hunt(version);
#endif
	if (!i) {
		printk("pcd: No Panasonic CD-ROM device found.\n");
#ifdef MODULE
		return -EIO;
#else
		return kmem_start;
#endif
	}
	if(register_blkdev(MAJOR_NR, "pcd", &pcd_fops)) {
		printk("pcd: Unable to get major %d for CD-ROM driver.\n", MAJOR_NR);
#ifdef MODULE
		return -EIO;
#else
		return kmem_start;
#endif
	}
#ifdef MODULE
	printk("pcd: %s, addr = %03x, %s interface; module inserted.\n", version, pcd.base, pcd.if_type == PCD_SB_IF ? "SB" : "normal");
#else
	printk("pcd: %s, addr = %03x, %s interface; driver installed.\n", version, pcd.base, pcd.if_type == PCD_SB_IF ? "SB" : "normal");
#endif
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	read_ahead[MAJOR_NR] = PCD_READ_AHEAD;
	request_region(pcd.base, 4, "pcd");
	pcd.present = 1;
	pcd.change_flag = 0;
	pcd.open_count = 0;
	pcd.paused = 0;
	pcd.eject_sw = 0;
	blk_size[MAJOR_NR] = pcd_blk_size;
	blksize_size[MAJOR_NR] = pcd_blocksizes;
#ifdef MODULE
	return 0;
#else
	return kmem_start;
#endif
}	

#ifdef MODULE
void cleanup_module(void)
{
	if (MOD_IN_USE) {
		printk("pcd: Module in use, cannot remove.\n");
		return;
	}
	release_region(pcd.base, 4);
	if (unregister_blkdev(MAJOR_NR, "pcd") == -EINVAL) {
		printk("pcd: Cannot unregister block device.\n");
		return;
	}
	blk_size[MAJOR_NR] = NULL;
	printk("pcd: Module removed.\n");
}
#endif

/*
 * read in data block into the buffer if not present
 * and set pcd.bufaddr
 */
static inline int pcd_readblock(char **bufp, int block)
{
	u_char m, s, f;
	int e, n;

	/* if block is not in the buffer...  */
	if (block < pcd.bufaddr || block >= pcd.bufaddr + PCD_BUFFER_SIZE) {
#ifdef PCD_STAT
		++stat.readcachemisses;
#endif
		block2msf(&m, &s, &f, block);
		/* if read ahead runs off trim n */
		if (block + PCD_BUFFER_SIZE >= pcd.toc.datasize)
			n = pcd.toc.datasize - block;
		else
			n = PCD_BUFFER_SIZE;
		e = pcd_readcmd(m, s, f, pcd.buffer, n);
		pcd.bufaddr = e ? -PCD_BUFFER_SIZE : block;
		if (e)
			return e;
	}
#ifdef PCD_STAT
	else
		++stat.readcachehits;
#endif
	*bufp = pcd.buffer + (block - pcd.bufaddr) * CDBLOCK_SIZE;
	return 0;
}

/*
 * read nsect sectors into the buf
 */
static inline int pcd_read(unsigned nsect, unsigned sectaddr, char *buf)
{
	int block, startblock, endblock;
	int startoffset, endoffset, transfersize;
	int e;
	char *buffer;

	/*
	 * note that we have to cope with the
	 * different block sizes between the kernel and the device
	 */
#ifdef PCD_STAT
	++stat.totalreads;
#endif
	startblock = sectaddr >> 2;
	endblock = (sectaddr + nsect - 1) >> 2;
	if (endblock >= pcd.toc.datasize)
		return 0;
	for (block = startblock; block <= endblock; ++block) {
		e = pcd_readblock(&buffer, block);
		if (e) {
			perr(e);
			return 0;
		}
		/* start and end offset within the block */
		startoffset = block == startblock ?
			      sectaddr & 3 :
			      0;
		endoffset = block == endblock ?
		            ((sectaddr + nsect - 1) & 3) + 1:
			    4;
		transfersize = (endoffset - startoffset) * SECTOR_SIZE;
		memcpy(buf, buffer + startoffset * SECTOR_SIZE, transfersize);
		buf += transfersize;
	}
	return 1;
}

/*
 * strategy routine called by the kernel
 *
 * for now it's do_sbpcd_request since we don't want kernel patches...
 */
static void DEVICE_REQUEST(void)
{
	int r;

	for (;;) {
		INIT_REQUEST;
		sti();
		switch(CURRENT->cmd) {
		case READ:
			if (pcd.changed) {
				printk("pcd: CD changed, I/O disabled until device is closed.\n");
				end_request(0);
				continue;
			}
			r = pcd_read(CURRENT->nr_sectors, CURRENT->sector, CURRENT->buffer);
			end_request(r);
			break;
		case WRITE:
			end_request(0);
			break;
		default:
			panic("pcd: Unknown command in do_pcd_request");
			break;
		}
	}
}

#if 0
/*
 * the cleaner but slower delay routines
 */
static void pcd_timerproc(unsigned long data)
{
	wake_up(&pcd_waitqueue);
}

static void pcd_delay(unsigned int t)
{
	if (t < 1)
		t = 1;
	pcd_timer.expires = t;
	add_timer(&pcd_timer);
	sleep_on(&pcd_waitqueue);
	del_timer(&pcd_timer);
}
#endif

/* delay t jiffies */
static void pcd_delay(unsigned int t)
{
	unsigned long flags;
	unsigned long to;
#ifdef PCD_STAT
	unsigned i = 0;
#endif

	to = jiffies + t;
	while (jiffies < to) {
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = to;
		save_flags(flags);
		sti();
		schedule();
		restore_flags(flags);
#ifdef PCD_STAT
		++i;
#endif
	}
#ifdef PCD_STAT
		if (i > stat.delayloop)
			stat.delayloop = i;
#endif
}

/* wait for given flags in the pcd.ready port */
static int wait_data(u_char mask, int sleep_to)
{
	int i;
#ifdef PCD_STAT
	unsigned start;
#endif

	/* first fast busy loop... */
	for (i = 0; i < BUSY_TO; ++i) {
		if (((pcd.ready_value = INB(pcd.ready)) & mask) != mask) {
#ifdef PCD_STAT
			++stat.busyloop[i];
#endif
			return OK;
		}
	}
	/* ... then slower with delays */
#ifdef PCD_STAT
	start = jiffies;
#endif
	for (i = 0; i < sleep_to; ++i) {
		if (((pcd.ready_value = INB(pcd.ready)) & mask) != mask) {
#ifdef PCD_STAT
			stat.sleeptime[i] = jiffies - start;
			++stat.sleeploop[i];
#endif
			return OK;
		}
		pcd_delay(1);
	}
#ifdef PCD_STAT
	if (stat.timeout_index < MAXTIMEOUT)
		stat.timeout[stat.timeout_index++] = jiffies - start;
#endif
	return TIMEOUT;
}

int pcd_open(struct inode *ip, struct file *fp)
{
	int e, i;

	if (!pcd.present)
		return -ENXIO;
	if (fp->f_mode & 2)
		return -EROFS;
	if (pcd.open_count)
		return -EBUSY;
#ifdef CLOSE_TRAY_ON_OPEN
	e = pcd_getstat();
	if (!(pcd.status & STAT_TRAY)) {
		e = pcd_close(1);
		if (e) {
			perr(e);
			return -EIO;
		}
	}
#endif
	/* let's retry here -- maybe the tray is just closed */
	for (i = 0; i < OPEN_RETRY_N; ++i) {
		e = pcd_getstatd();
		/* media changed error msg is OK */
		if (e == ERR_DISKCHANGED)
			e = 0;
		if (e && OPEN_RETRYABLE(e)) {
			pcd_delay(OPEN_DELAY);
			continue;
		}
	}
	if (e) {
		perr(e);
		return -EIO;
	}
#ifdef PCD_LOCK
	PCD_CMD_C(lock, e);
#endif
	e = pcd_readtoc();
	if (e) {
		perr(e);
#ifdef PCD_LOCK
		/* try to unlock without error handling */
		(void)pcd_unlock();
#endif
		return -EIO;
	}
	/*pcd.change_flag = pcd.changed;*/
#ifndef INVALIDATE_BUF_ON_CLOSE
	if (pcd.changed)
		invalidate_buffers(ip->i_rdev);
#endif
	pcd.changed = 0;
	MOD_INC_USE_COUNT;
	++pcd.open_count;
	pcd.bufaddr = -PCD_BUFFER_SIZE;
#ifdef PCD_STAT
	memset(&stat, 0, sizeof stat);
#endif
	return 0;
}

#ifdef PCD_STAT
void printstats(void)
{
	int i, j, s, inc;

	printk("pcd: Statistics:\n");
	printk("pcd:   totalreads = %u, readcachehits = %u, readcachemisses = %u\n",
	       stat.totalreads, stat.readcachehits, stat.readcachemisses);
	for (i = 0, inc = BUSY_TO / 16 > 0 ? BUSY_TO / 16 : 1; i < BUSY_TO; i += inc) {
		for (j = i, s = 0; j < i + inc; ++j)
			s += stat.busyloop[j];
		if (s)
			printk("pcd:   busyloop[%d - %d] = %d\n", i, i + inc - 1, s);
	}
	for (i = 0; i < DISK_SLEEP_TO; i += 1) {
		s = stat.sleeploop[i];
		if (s) {
			printk("pcd:   sleeploop[%d] = %d, ", i, s);
			printk("sleeptime[%d] = %d\n", i, stat.sleeptime[i]);
		}
	}
	for (i = 0; i < RETRY_N; i += 1) {
		s += stat.cmdretry[i];
		if (s)
			printk("pcd:   cmdretry[%d] = %d\n", i, s);
	}
	for (i = 0; i < stat.timeout_index; i += 1) {
		s = stat.timeout[i];
		printk("pcd:   timeout[%d] = %d\n", i, s);
	}
	printk("pcd:   delayloop = %d\n", stat.delayloop);
}
#endif

void pcd_release(struct inode *ip, struct file *filep)
{
	int e;

	pcd.open_count = 0;
	MOD_DEC_USE_COUNT;
#ifdef INVALIDATE_BUF_ON_CLOSE
	invalidate_buffers(ip->i_rdev);
#endif
#ifdef PCD_LOCK
	e = pcd_unlock();
#endif
	if (e) {
		perr(e);
		return;
	}
	if (pcd.eject_sw) {
		e = pcd_eject();
		if (e) {
			perr(e);
			return;
		}
	}
#ifdef PCD_STAT
	printstats();
#endif
}

/*
 * is disk changed?
 */
int check_pcd_change(dev_t dev)
{
#ifdef PCD_DEBUG
	printk("pcd: pcd.change_flag = %d\n", pcd.change_flag);
#endif
	if (pcd.change_flag) {
		pcd.change_flag = 0;
		return 1;
	}
	return 0;
}

int pcd_ioctl(struct inode *inodep, struct file *fp, unsigned int cmd, unsigned long arg)
{
	int e, track;
	u_char vol;
	struct cdrom_ti ti;
	struct cdrom_msf msf;
	struct cdrom_tochdr tochdr;
	struct cdrom_tocentry tocentry;
	struct cdrom_volctrl volctrl;
	struct cdrom_subchnl subchnl;
	struct cdrom_multisession multisession;

	if (inodep == NULL)
		return -EINVAL;
	if (!pcd.present)
		return -ENXIO;
	/*
	 * read status and update toc if necessary
	 */
	e = pcd_getstat();
	/* media changed msg is OK */
	if (e && e != ERR_DISKCHANGED) {
		perr(e);
		return -EIO;
	}
	if (pcd.update_toc)
		PCD_CMD_C(readtoc, e)

#ifdef PCD_LOCK
	if (pcd.unlocked && cmd != CDROMEJECT) {
		pcd.unlocked = 0;
		(void)pcd_lock();
	}
#endif

	switch (cmd) {
	case CDROMPAUSE:	/* pause audio playing */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMPAUSE\n");
#endif
		PCD_CMD_C(pause, e)
		break;
	case CDROMRESUME:	/* resume playing after a pause */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMRESUME\n");
#endif
		PCD_CMD_C(resume, e)
		break;
	case CDROMPLAYMSF:	/* play range given in m-s-f */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMPLAYMSF\n");
#endif
		e = verify_area(VERIFY_READ, (void *)arg, sizeof(msf));
		if (e)
			return e;
		memcpy_fromfs(&msf, (void *)arg, sizeof(msf));
#ifdef PCD_DEBUG
		printk("pcd: from m/s/f = %d/%d/%d, to m/s/f = %d/%d/%d\n",
		       msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0,
		       msf.cdmsf_min1, msf.cdmsf_sec1, msf.cdmsf_frame1);
#endif
		if (msf_less(pcd.toc.size.m, pcd.toc.size.s, pcd.toc.size.f,
		             msf.cdmsf_min1, msf.cdmsf_sec1, msf.cdmsf_frame1))
			return -EINVAL;
		e = pcd_playmsf(msf.cdmsf_min0, msf.cdmsf_sec0, msf.cdmsf_frame0,
		                msf.cdmsf_min1, msf.cdmsf_sec1, msf.cdmsf_frame1);
		if (e) {
			perr(e);
			return -EIO;
		}
		break;
	case CDROMPLAYTRKIND:	/* play tracks (does _not_ ignore index) */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMPLAYTRKIND\n");
#endif
		e = verify_area(VERIFY_READ, (void *)arg, sizeof(ti));
		if (e)
			return e;
		memcpy_fromfs(&ti, (void *)arg, sizeof(ti));
		if (ti.cdti_trk0 < pcd.toc.start_track ||
		    ti.cdti_trk1 > pcd.toc.end_track ||
		    ti.cdti_trk0 > ti.cdti_trk1)
			return -EINVAL;
		e = pcd_playti(ti.cdti_trk0, ti.cdti_ind0, ti.cdti_trk1, ti.cdti_ind1);
		if (e) {
			perr(e);
			return -EIO;
		}
		break;
	case CDROMREADTOCHDR:	/* read toc header */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMREADTOCHDR\n");
#endif
		e = verify_area(VERIFY_WRITE, (void *)arg, sizeof(tochdr));
		if (e)
			return e;
		tochdr.cdth_trk0 = pcd.toc.start_track;
		tochdr.cdth_trk1 = pcd.toc.end_track;
		memcpy_tofs((void *)arg, &tochdr, sizeof(tochdr));
		break;
	case CDROMREADTOCENTRY:	/* read toc entry */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMREADTOCENTRY\n");
#endif
		e = verify_area(VERIFY_WRITE, (void *)arg, sizeof(tocentry));
		if (e)
			return e;
		memcpy_fromfs(&tocentry, (void *)arg, sizeof(tocentry));
#ifdef PCD_DEBUG
		printk("pcd: track: %d, format: %d\n", 
		       tocentry.cdte_track, tocentry.cdte_format);
#endif
		track = tocentry.cdte_track == CDROM_LEADOUT ? pcd.toc.end_track + 1 : tocentry.cdte_track;
		if (track < pcd.toc.start_track || track > pcd.toc.end_track + 1)
			return -EINVAL;
		tocentry.cdte_adr = pcd.toc.entry[track].adr;
		tocentry.cdte_ctrl = pcd.toc.entry[track].ctrl;
		tocentry.cdte_datamode = pcd.toc.entry[track].datamode;
		switch (tocentry.cdte_format) {
		case CDROM_LBA:
			tocentry.cdte_addr.lba = msf2block(pcd.toc.entry[track].start.m,
			                                   pcd.toc.entry[track].start.s,
			                                   pcd.toc.entry[track].start.f);
			break;
		case CDROM_MSF:
			tocentry.cdte_addr.msf.minute = pcd.toc.entry[track].start.m;
			tocentry.cdte_addr.msf.second = pcd.toc.entry[track].start.s;
			tocentry.cdte_addr.msf.frame = pcd.toc.entry[track].start.f;
			break;
		default:
			return -EINVAL;
		}
		memcpy_tofs((void *)arg, &tocentry, sizeof(tocentry));
		break;
	case CDROMSTOP:		/* stop audio playing */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMSTOP\n");
#endif
		PCD_CMD_C(stop, e)
		break;
	case CDROMSTART:	/* start the drive (does nothing) */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMSTART\n");
#endif
		/* empty */
		break;
	case CDROMEJECT:	/* eject tray */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMEJECT\n");
#endif
		PCD_CMD_C(unlock, e);
		pcd.unlocked = 1;
		PCD_CMD_C(eject, e)
		break;
	case CDROMVOLCTRL:	/* set volume */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMVOLCTRL\n");
#endif
		e = verify_area(VERIFY_READ, (void *)arg, sizeof(volctrl));
		if (e)
			return e;
		memcpy_fromfs(&volctrl, (void *)arg, sizeof(volctrl));
		if (volctrl.channel0 > volctrl.channel1)
			vol = volctrl.channel0;
		else
			vol = volctrl.channel1;
		e = pcd_volume(vol);
		if (e) {
			perr(e);
			return -EIO;
		}
		break;
	case CDROMSUBCHNL:	/* give status info */
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMSUBCHNL\n");
#endif
		e = verify_area(VERIFY_WRITE, (void *)arg, sizeof(subchnl));
		if (e)
			return e;
		memcpy_fromfs(&subchnl, (void *)arg, sizeof(subchnl));
		subchnl.cdsc_audiostatus = pcd.status & STAT_PLAY ? 
		                           pcd.paused ? CDROM_AUDIO_PAUSED : CDROM_AUDIO_PLAY :
					   pcd.prev_status & STAT_PLAY ? CDROM_AUDIO_COMPLETED : CDROM_AUDIO_NO_STATUS;
		e = pcd_readsubq();
		/* 
		 * don't write error messages: this ioctl is used to be
		 * called to poll CDROM status in CD players
		 */
		if (e)
			return -EIO;
		subchnl.cdsc_adr = pcd.subq.adr;
		subchnl.cdsc_ctrl = pcd.subq.ctrl;
		subchnl.cdsc_trk = pcd.subq.track;
		subchnl.cdsc_ind = pcd.subq.index;
		switch (subchnl.cdsc_format) {
		case CDROM_LBA:
			subchnl.cdsc_absaddr.lba = msf2block(pcd.subq.absaddr.m,
			                                     pcd.subq.absaddr.s,
							     pcd.subq.absaddr.f);
			subchnl.cdsc_reladdr.lba = msf2block(pcd.subq.reladdr.m,
			                                     pcd.subq.reladdr.s,
							     pcd.subq.reladdr.f);
			break;
		case CDROM_MSF:
			subchnl.cdsc_absaddr.msf.minute = pcd.subq.absaddr.m;
			subchnl.cdsc_absaddr.msf.second = pcd.subq.absaddr.s;
			subchnl.cdsc_absaddr.msf.frame = pcd.subq.absaddr.f;
			subchnl.cdsc_reladdr.msf.minute = pcd.subq.reladdr.m;
			subchnl.cdsc_reladdr.msf.second = pcd.subq.reladdr.s;
			subchnl.cdsc_reladdr.msf.frame = pcd.subq.reladdr.f;
			break;
		default:
			return -EINVAL;
		}
		memcpy_tofs((void *)arg, &subchnl, sizeof(subchnl));
		break;
	case CDROMREADMODE2:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMREADMODE2\n");
#endif
		break;
	case CDROMREADMODE1:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMREADMODE1\n");
#endif
		break;
#ifndef OLD
	case CDROMEJECT_SW:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMEJECT_SW\n");
#endif
		pcd.eject_sw = arg;
		break;
#endif
	case CDROMMULTISESSION:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMMULTISESSION\n");
#endif
		e = verify_area(VERIFY_WRITE, (void *)arg, sizeof(multisession));
		if (e)
			return e;
		memcpy_fromfs(&multisession, (void *)arg, sizeof(multisession));
		multisession.xa_flag = pcd.multi.flag;
		if (pcd.multi.flag)
			switch (multisession.addr_format) {
			case CDROM_LBA:
				multisession.addr.lba = msf2block(pcd.multi.base.m,
				                                  pcd.multi.base.s,
								  pcd.multi.base.f);
				break;
			case CDROM_MSF:
				multisession.addr.msf.minute = pcd.multi.base.m;
				multisession.addr.msf.second = pcd.multi.base.s;
				multisession.addr.msf.frame = pcd.multi.base.f;
				break;
			default:
				return -EINVAL;
			}
		memcpy_tofs((void *)arg, &multisession, sizeof(multisession));
		break;
#ifndef OLD
	case CDROMRESET:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMRESET\n");
#endif
		e = pcd_reset();
		if (e == 0)
			return -EIO;
		/*
		 * read status and update toc if necessary
		 */
		e = pcd_getstat();
		/* media changed msg is OK */
		if (e && e != ERR_DISKCHANGED) {
			perr(e);
			return -EIO;
		}
		if (pcd.update_toc)
			PCD_CMD_C(readtoc, e)
#ifdef PCD_LOCK
		pcd.unlocked = 0;
		(void)pcd_lock();
#endif
		break;
	case CDROMVOLREAD:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMVOLREAD\n");
#endif
		e = verify_area(VERIFY_WRITE, (void *)arg, sizeof(volctrl));
		if (e)
			return e;
		e = pcd_readvolume(&vol);
		if (e) {
			perr(e);
			return -EIO;
		}
		volctrl.channel0 = volctrl.channel1 = vol;
		volctrl.channel2 = volctrl.channel3 = 0;
		memcpy_tofs((void *)arg, &volctrl, sizeof(volctrl));
		break;
	case CDROMCLOSETRAY:
#ifdef PCD_DEBUG
		printk("pcd: ioctl CDROMCLOSETRAY\n");
#endif
		pcd_close(0);
		break;
#endif
	default:
		printk("pcd: Got unknown ioctl call: 0x%x\n", cmd);
		return -EINVAL;
	}
	return 0;
}

